

#include <string.h>
#include "discoord.h"


#define WGS84_MAJOR_AXIS 6378137.0000 // meters
#define WGS84_MINOR_AXIS 6356752.3142 // meters

// xyz converters
vector4d local2dismx[4];
vector4d dis2localmx[4];

// local dis/world dis rotations
vector4d center2worldmxv[4];
vector4d world2centermxv[4];

// local origin/world origin rotations
vector4d local2dismxv[4];
vector4d dis2localmxv[4];

int dis_exercise_id = 1;
int dis_site_id = 1;
int dis_application_id = 1;


/* **********************************************************
********************************************************** */
void set_dis_interface(int exercise_id, int site_id, int app_id) {

   dis_exercise_id = exercise_id ? exercise_id : 1;

   switch (site_id) {

      case 0:
         dis_site_id = 1;
         break;

      case 65535:
      case 65534:
         dis_site_id = 65534;
         break;

      default:
         dis_site_id = site_id;
         break;
   }

   switch (app_id) {

      case 0:
         dis_application_id = 1;
         break;

      case 65535:
      case 65534:
         dis_application_id = 65534;
         break;

      default:
         dis_application_id = app_id;
         break;
   }

}


/* **********************************************************
********************************************************** */
void get_dis_interface(int *exercise_id, int *site_id, int *app_id) {

   *exercise_id = dis_exercise_id;
   *site_id = dis_site_id;
   *app_id = dis_application_id;
}


/* **********************************************************
********************************************************** */
void dislocal2disworld(lla_type *center) {

   init_mx(center2worldmxv);
   rotate_mx_yd(center2worldmxv, -center->latitude);
   rotate_mx_zd(center2worldmxv, center->longitude);
   inversemxrtd(center2worldmxv, world2centermxv);

   matmatmultod(center2worldmxv, local2dismx); 
   local2dismx[0][3] = local2dismx[0][2] * WGS84_MAJOR_AXIS;
   local2dismx[1][3] = local2dismx[1][2] * WGS84_MAJOR_AXIS;
   local2dismx[2][3] = local2dismx[2][2] * WGS84_MINOR_AXIS;

   inversemxrtd(local2dismx, dis2localmx);

   inversemxrtd(local2dismxv, dis2localmxv);
}


/* **********************************************************
********************************************************** */
void init_disvissim(lla_type *center) {

   init_mx(local2dismx);
   rotate_mx_zd(local2dismx, HALFPI);
   rotate_mx_yd(local2dismx, HALFPI);

   init_mx(local2dismxv);
   rotate_mx_yd(local2dismxv, PI);
   rotate_mx_zd(local2dismxv, -HALFPI);

   dislocal2disworld(center);
}


/* **********************************************************
********************************************************** */
void local2dis(vector3d local, vector3d dis, int type) {

   if (type == DIS_POSITION) {
      dis[0] = dotproduct3(local2dismx[0], local) + local2dismx[0][3];
      dis[1] = dotproduct3(local2dismx[1], local) + local2dismx[1][3];
      dis[2] = dotproduct3(local2dismx[2], local) + local2dismx[2][3];
      return;
   }
   
   dis[0] = dotproduct3(local2dismx[0], local);
   dis[1] = dotproduct3(local2dismx[1], local);
   dis[2] = dotproduct3(local2dismx[2], local);   
}


/* **********************************************************
********************************************************** */
void dis2local(vector3d dis, vector3d local, int type) {

   if (type == DIS_POSITION) {
      local[0] = dotproduct3(dis2localmx[0], dis) + dis2localmx[0][3];
      local[1] = dotproduct3(dis2localmx[1], dis) + dis2localmx[1][3];
      local[2] = dotproduct3(dis2localmx[2], dis) + dis2localmx[2][3];
      return;
   }

   local[0] = dotproduct3(dis2localmx[0], dis);
   local[1] = dotproduct3(dis2localmx[1], dis);
   local[2] = dotproduct3(dis2localmx[2], dis);
}


/* **********************************************************
********************************************************** */
void local2disypr(vector3d localypr, vector3d disypr) {

   vector4d mx[4];
   double y, p, r;

//   init_mx(mx);
//   matmatmultvd(dis2localmxv, mx);

   copymx4x4o(mx, dis2localmxv);

   rotate_mx_yd(mx, localypr[2]);
   rotate_mx_xd(mx, localypr[1]);
   rotate_mx_zd(mx, localypr[0]);

   matmatmultvd(local2dismxv, mx);

   rotate_mx_yd(mx, -HALFPI);
   matmatmultvd(center2worldmxv, mx);

   decompose_zyypxrd(mx, &r, &p, &y);
   disypr[0] = y;
   disypr[1] = p;
   disypr[2] = r;
}


/* **********************************************************
********************************************************** */
void dis2localypr(vector3d disypr, vector3d localypr) {

   vector4d mx[4];
   double y, p, r;
   
   copymx4x4o(mx, local2dismxv);

   rotate_mx_xd(mx, disypr[2]);
   rotate_mx_yd(mx, disypr[1]);
   rotate_mx_zd(mx, disypr[0]);

   matmatmultvd(world2centermxv, mx);
   rotate_mx_yd(mx, HALFPI);

   matmatmultvd(dis2localmxv, mx);

   decompose_zyxpyrd(mx, &r, &p, &y);
   localypr[0] = y;
   localypr[1] = p;
   localypr[2] = r;
}


/* **********************************************************
   computes DIS time - cycles over 1 hour
********************************************************** */
unsigned int distime(double *seconds) {

   double t;

   t = *seconds /3600.0;
   t = t - (int)t;  // asdf windoze's round down - may have to alter for unix boxes
   t *= (1<<30) - 1;
   return ((unsigned int)t) & 0xfffffffe;
}


/* **********************************************************
   map geodetic (lat,lon, alt) to WGS 84
********************************************************** */
void lla2dis(lla_type *in, vector3d out) {

   vector3d w;
   double ad;
   double e2;
   double sinlat;
   double clat;
   double t;

   if (in->latitude > HALFPI-CORRECT) {
      out[0] = 0;
      out[1] = 0;
      out[2] = WGS84_MINOR_AXIS;
      return;
   }
   
   if (in->latitude < -HALFPI+CORRECT) {
      out[0] = 0;
      out[1] = 0;
      out[2] = -WGS84_MINOR_AXIS;
      return;
   }
   
   e2 = WGS84_MINOR_AXIS/WGS84_MAJOR_AXIS;
   e2 = 1-e2*e2;
   sinlat = sin(in->latitude);

#ifdef __ALGEBRAIC

   vector4d mx[4];
   vector3d v = { 1.0, 0.0, 0.0 };

   init_mx(mx);
   rotate_mx_y(mx, -in->latitude);
   rotate_mx_z(mx, in->longitude);

   w[0] = dotproduct3(mx[0], v);
   w[1] = dotproduct3(mx[1], v);
   w[2] = dotproduct3(mx[2], v);

#else

   clat = cos(in->latitude);
   w[0] = clat*cos(in->longitude);
   w[1] = clat*sin(in->longitude);
   w[2] = sin(in->latitude);

#endif

#ifdef __TRUE_ELLIPSOID
   t= (WGS84_MAJOR_AXIS + in->altitude);
   out[0] = w[0] * t;
   out[1] = w[1] * t;
   out[2] = w[2] * (WGS84_MINOR_AXIS + in->altitude);
#else   
   ad = WGS84_MAJOR_AXIS/sqrt(1 - e2*(sinlat*sinlat));
   t= (ad + in->altitude);
   out[0] = w[0] * t;
   out[1] = w[1] * t;
   out[2] = w[2] * (ad*(1-e2) + in->altitude);
#endif
}


/* **********************************************************
   map WGS 84 to geodetic (lat,lon, alt)
********************************************************** */
void dis2lla(vector3d in, lla_type *out) {

   double eprime2;
   double a2, b2;
   double p;
   double prime;
   double div, num;
   double e2;
   double cosprime, sinprime;
   double ad;
   double sinlat;
   
   if (fabs(in[0]) < 0)
      if (in[1] > 0)
         out->longitude = HALFPI;
      else
         out->longitude = -HALFPI;
   else
      out->longitude = atan(in[1]/in[0]);

   a2 = WGS84_MAJOR_AXIS*WGS84_MAJOR_AXIS;
   b2 = WGS84_MINOR_AXIS*WGS84_MINOR_AXIS;
   eprime2 = a2/b2 - 1.0;
   e2 = 1.0 - b2/a2;

   p = dotproduct2(in, in);
   p = p > CORRECT ? sqrt(p) : 0;
   
   div = WGS84_MINOR_AXIS*p;
   
   if (fabs(div) < 0)
      if (in[2] > 0)
         prime = HALFPI;
      else
         prime = -HALFPI;
   else
      prime = atan((WGS84_MAJOR_AXIS*in[2])/div);

   cosprime = cos(prime);
   sinprime = sin(prime);
   
   div = p - WGS84_MAJOR_AXIS*e2*cosprime*cosprime*cosprime;
   num = in[2] + WGS84_MINOR_AXIS*eprime2*sinprime*sinprime*sinprime;

   if (fabs(div) < 0)
      if (num > 0)
         out->latitude = HALFPI;
      else
	 out->latitude = -HALFPI;
   else
      prime = atan(num/div);

   sinlat = sin(out->latitude);

   ad = WGS84_MAJOR_AXIS*sqrt(1 - e2*(sinlat*sinlat));
   
   out->altitude = p*cos(out->latitude) + in[2]*sinlat - ad;
}


/* **********************************************************
********************************************************** */
DIS_header_type::DIS_header_type() {

   version = DIS_PDU_VERSION;
   exercise_id = dis_exercise_id;
}


/* **********************************************************
********************************************************** */
void DIS_header_type::build(sfile *sbuffer) {

   sbuffer->sput_char(version);
   sbuffer->sput_char(exercise_id);
   sbuffer->sput_char(type);
   sbuffer->sput_char(family);
   sbuffer->sput_uint(time);
   sbuffer->sput_ushort(size);
   sbuffer->sput_ushort(padd);
}


/* **********************************************************
********************************************************** */
void DIS_header_type::extract(sfile *sbuffer) {

   version = sbuffer->scan_char();
   exercise_id = sbuffer->scan_char();
   type = sbuffer->scan_char();
   family = sbuffer->scan_char();
   time = sbuffer->scan_uint();
   size = sbuffer->scan_ushort();
   sbuffer->skip(2);
}


/* **********************************************************
********************************************************** */
void DIS_entity_type_type::build(sfile *sbuffer) {

   sbuffer->sput_char(kind);
   sbuffer->sput_char(domain);
   sbuffer->sput_ushort(country);
   sbuffer->sput_char(category);
   sbuffer->sput_char(subcategory);
   sbuffer->sput_char(specific);
   sbuffer->sput_char(extra);
}


/* **********************************************************
********************************************************** */
void DIS_entity_type_type::extract(sfile *sbuffer) {

   kind = sbuffer->scan_char();
   domain = sbuffer->scan_char();
   country = sbuffer->scan_ushort();
   category = sbuffer->scan_char();
   subcategory = sbuffer->scan_char();
   specific = sbuffer->scan_char();
   extra = sbuffer->scan_char();
}


/* **********************************************************
********************************************************** */
void DIS_burst_descriptor_type::build(sfile *sbuffer) {

   burst_id.build(sbuffer);
   sbuffer->sput_ushort(burst_warhead);
   sbuffer->sput_ushort(burst_fuze);
   sbuffer->sput_ushort(burst_quantity);
   sbuffer->sput_ushort(burst_rate);
}


/* **********************************************************
********************************************************** */
void DIS_burst_descriptor_type::extract(sfile *sbuffer) {

   burst_id.extract(sbuffer);

   burst_warhead = sbuffer->scan_ushort();
   burst_fuze = sbuffer->scan_ushort();
   burst_quantity = sbuffer->scan_ushort();
   burst_rate = sbuffer->scan_ushort();
}
